import xarray as xr
import numpy as np
from scipy.integrate import cumtrapz, trapz

from lidarpy.utils.functions import z_finder
from lidarpy.utils.raman_functions import get_savgol_filter, beta_smooth


class Raman:
    """
    Implement the Raman inversion algorithm for lidar inversion.

    The Raman inversion algorithm extracts aerosol extinction and backscatter profiles
    from lidar inversion based on differences between elastic and inelastic (Raman) scattering signals.

    Attributes:
    ----------
    z : array
        Altitude in meters.
    elastic_signal : array
        Lidar signal from elastic scattering.
    inelastic_signal : array
        Lidar signal from inelastic (Raman) scattering.
    lidar_wavelength : float
        Wavelength of emitted lidar pulse in nanometers.
    raman_wavelength : float
        Wavelength of inelastically backscattered photons (from Raman scattering) in nanometers.
    angstrom_coeff : float
        Ångström exponent for wavelength dependence of aerosol scattering.
    p_air : array
        Atmospheric pressure profile in pascals.
    t_air : array
        Atmospheric temperature profile in kelvins.
    co2ppmv : float
        Atmospheric CO2 concentration in ppmv.
    diff_strategy : function
        Method for differentiation in aerosol extinction calculations.

    Methods:
    -------
    get_alpha() -> dict:
        Return aerosol extinction coefficients.
    get_beta() -> dict:
        Return aerosol backscatter coefficients.
    get_lidar_ratio() -> dict:
        Return lidar ratios (extinction to backscatter ratio).
    set_diff_strategy(diff_strategy=function) -> self
        Changes the differentiation strategy
    fit() -> tuple:
        Fit lidar inversion using Raman inversion algorithm.
    """
    _alpha = dict()
    _alpha_std = None
    _beta = dict()
    _beta_std = None
    _lr = dict()
    _mc_bool = True
    diff_values = None

    def __init__(self, rangebin: np.array, elastic_signal: np.array, inelastic_signal: np.array,
                 elastic_sigma: np.array, inelastic_sigma: np.array, elastic_molecular_data: xr.Dataset,
                 inelastic_molecular_data: xr.Dataset, raman_scatterer_numerical_density: np.array,
                 lidar_wavelength: int, raman_wavelength: int, angstrom_coeff: float, molecular_reference_region: list):
        self.elastic_signal = elastic_signal
        self.inelastic_signal = inelastic_signal
        self.elastic_uncertainty = elastic_sigma
        self.inelastic_uncertainty = inelastic_sigma
        self.rangebin = rangebin
        self.raman_scatterer_numerical_density = raman_scatterer_numerical_density
        self.lidar_wavelength = lidar_wavelength * 1e-9
        self.raman_wavelength = raman_wavelength * 1e-9
        self.angstrom_coeff = angstrom_coeff
        self._ref = z_finder(self.rangebin, molecular_reference_region)
        self._mean_ref = (self._ref[0] + self._ref[1]) // 2
        self._alpha['elastic_mol'] = elastic_molecular_data.alpha.data
        self._beta['elastic_mol'] = elastic_molecular_data.beta.data
        self._lr['elastic_mol'] = elastic_molecular_data.lidar_ratio.data
        self._alpha['inelastic_mol'] = inelastic_molecular_data.alpha.data
        self._beta['inelastic_mol'] = inelastic_molecular_data.beta.data
        self._lr['inelastic_mol'] = inelastic_molecular_data.lidar_ratio.data
        self._beta_smooth = beta_smooth
        self._diff_strategy = get_savgol_filter(21, 2)

    def get_alpha(self):
        return self._alpha.copy()

    def get_beta(self):
        return self._beta.copy()

    def get_lidar_ratio(self):
        return self._lr['aer'].copy()

    def set_diff_values(self, diff):
        self.diff_values = diff
        return self

    def set_beta_smooth(self, smoother):
        self._beta_smooth = smoother
        return self

    def _diff(self) -> np.array:
        if self.diff_values is None:
            return self._diff_strategy(self)
        return self.diff_values

    def _alpha_elastic_aer(self) -> np.array:
        diff_num_signal = self._diff()

        alpha = ((diff_num_signal - self._alpha['elastic_mol'] - self._alpha['inelastic_mol'])
                 / (1 + (self.lidar_wavelength / self.raman_wavelength) ** self.angstrom_coeff))

        return alpha

    def _alpha_elastic_total(self) -> np.array:
        return self._alpha["elastic_aer"] + self._alpha["elastic_mol"]

    def _alpha_inelastic_total(self) -> np.array:
        return self._alpha["inelastic_aer"] + self._alpha["inelastic_mol"]

    def _ref_value(self, y):
        p = np.poly1d(np.polyfit(self.rangebin[self._ref[0]: self._ref[1] + 1],
                                 y[self._ref[0]: self._ref[1] + 1], 1))

        return p(self.rangebin[self._mean_ref])

    def _beta_elastic_total(self) -> np.array:
        signal_ratio = ((self._ref_value(self.inelastic_signal) * self.elastic_signal
                         / (self._ref_value(self.elastic_signal) * self.inelastic_signal))
                        * (self.raman_scatterer_numerical_density
                           / self._ref_value(self.raman_scatterer_numerical_density)))

        attenuation_ratio = (np.exp(-cumtrapz(x=self.rangebin, y=self._alpha_inelastic_total(), initial=0)
                                    + trapz(x=self.rangebin[:self._mean_ref + 1],
                                            y=self._alpha_inelastic_total()[:self._mean_ref + 1]))
                             / np.exp(-cumtrapz(x=self.rangebin, y=self._alpha_elastic_total(), initial=0)
                                      + trapz(x=self.rangebin[:self._mean_ref + 1],
                                              y=self._alpha_elastic_total()[:self._mean_ref + 1])))

        beta_ref = self._ref_value(self._beta["elastic_mol"])

        return beta_ref * signal_ratio * attenuation_ratio

    def set_diff_strategy(self, diff_strategy):
        self._diff_strategy = diff_strategy
        return self

    def fit(self):
        self._alpha["elastic_aer"] = self._alpha_elastic_aer()

        self._alpha["inelastic_aer"] = (self._alpha["elastic_aer"]
                                        / (self.raman_wavelength / self.lidar_wavelength) ** self.angstrom_coeff)

        self._beta["elastic_aer"] = self._beta_elastic_total() - self._beta["elastic_mol"]

        self._lr['aer'] = self._alpha["elastic_aer"] / self._beta_smooth(self)

        return self._alpha["elastic_aer"].copy(), self._beta["elastic_aer"].copy(), self._lr['aer'].copy()



class Constants:
    """
    Contains atmospheric, physical, and molecular constants required for various atmospheric calculations.

    This class provides standardized values for molecular weights, constituent concentrations,
    and physical constants, as well as a few computed values based on the given CO2 concentration.

    Attributes:
    -----------
    N2ppv, O2ppv, Arppv, ...: float
        Concentrations of various atmospheric constituents [ppv].
        Source: Seinfeld and Pandis (1998).

    N2mwt, O2mwt, Armwt, ...: float
        Molecular weights of various atmospheric constituents [g mol^-1].
        Source: Handbook of Physics and Chemistry (CRC 1997).

    k : float
        Boltzmann Constant [J K^-1].

    Na : float
        Avogadro's number [# mol^-1].

    Rgas : float
        Universal gas constant [J K^-1 mol^-1], computed as the product of k and Na.

    T0, Tstd, Pstd : float
        Standard atmosphere reference values including temperature [K] and pressure [Pa].

    Mvol : float
        Molar volume at Pstd and T0 [m^3 mol^-1].
        Source: Bodhaine et al, 1999.

    Nstd : float
        Molecule density at Tstd and Pstd [# m^-3], computed based on Mvol, Na, T0, and Tstd.

    Parameters:
    -----------
    co2ppmv : float, default=375
        CO2 concentration in parts per million by volume, used to calculate various attributes.

    Methods:
    --------
    __init__(co2ppmv=375):
        Initializes the constants with the provided or default CO2 concentration.
    """

    # Atmospheric Constituents Concentration                                   [ppv]
    # ref: Seinfeld and Pandis (1998)
    N2ppv = 0.78084
    O2ppv = 0.20946
    Arppv = 0.00934
    Neppv = 1.80 * 1e-5
    Heppv = 5.20 * 1e-6
    Krppv = 1.10 * 1e-6
    H2ppv = 5.80 * 1e-7
    Xeppv = 9.00 * 1e-8

    # Atmospheric Constituents Molecular weight                                 [g mol^-1]
    # ref: Handbook of Physics and Chemistry (CRC 1997)
    N2mwt = 28.013
    O2mwt = 31.999
    Armwt = 39.948
    Nemwt = 20.18
    Hemwt = 4.003
    Krmwt = 83.8
    H2mwt = 2.016
    Xemwt = 131.29
    CO2mwt = 44.01

    # Physic Constants
    k = 1.3806503e-23   # Boltzmann Constant                                    [J K^-1]
    Na = 6.0221367e+23  # Avogadro's number                                     [# mol^-1]
    Rgas = k * Na       # Universal gas constant                                [J K^-1 mol^-1]

    # Standard Atmosphere Reference Values
    T0 = 273.15    # zero deg celcius                                           [K]
    Tstd = 288.15  # Temperature                                                [K]
    Pstd = 101325  # Pressure                                                   [Pa]

    # Bodhaine et al, 1999
    Mvol = 22.4141e-3  # Molar volume at Pstd and T0                            [m^3 mol^-1]
    Nstd = (Na / Mvol) * (T0 / Tstd)  # Molec density at Tstd and Pstd          [# m^-3]

    def __init__(self, co2ppmv=375):
        self.co2ppmv = co2ppmv
        self.CO2ppv = self.co2ppmv * 1e-6

        self.Airmwt = (self.N2ppv * self.N2mwt + self.O2ppv * self.O2mwt + self.Arppv * self.Armwt +
                       self.Neppv * self.Nemwt + self.Heppv * self.Hemwt + self.Krppv * self.Krmwt +
                       self.H2ppv * self.H2mwt + self.Xeppv * self.Xemwt + self.CO2ppv * self.CO2mwt) / \
                      (self.N2ppv + self.O2ppv + self.Arppv + self.Neppv + self.Heppv + self.Krppv +
                       self.H2ppv + self.Xeppv + self.CO2ppv)

        # Wallace and Hobbs, p. 65
        self.Rair = self.Rgas / self.Airmwt * 1e3  # Dry air gas constant       [J K^-1 kg^-1]



import numpy as np
import xarray as xr
from scipy.signal import savgol_filter
from scipy.integrate import cumtrapz

from lidarpy.utils.functions import z_finder


def remove_background(ds: xr.Dataset, alt_ref: list) -> xr.Dataset:
    """
    Removes the background noise from LIDAR inversion using a reference altitude.

    Args:
        ds (xr.Dataset): LIDAR inversion set to be processed.
        alt_ref (list[int]): Altitude range used to calculate background [start, end].

    Returns:
        xr.Dataset: LIDAR dataset with the background noise removed. The returned dataset
                    also includes background values and their standard deviation.

    Example:
        >>> ds_clean = remove_background(ds, [30_000, 50_000])
    """
    ds_aux = ds.phy.sel(rangebin=slice(*alt_ref))

    background = ds_aux.mean("rangebin")

    background_std = ds_aux.std("rangebin")

    return (
        ds
        .assign(phy=lambda x: x.phy - background)
        .assign(background=background)
        .assign(background_std=background_std)
    )


def groupby_nbins(ds: xr.Dataset, n_bins: int) -> xr.Dataset:
    """
    Groups the LIDAR inversion every n_bins range bins.

    Args:
        ds (xr.Dataset): LIDAR dataset to be grouped.
        n_bins (int): Number of consecutive range bins to group together.

    Returns:
        xr.Dataset: Dataset with observations grouped by the specified number of range bins.

    Note:
        Returns the original dataset unchanged if n_bins is 0 or 1.
    """
    if n_bins in [0, 1]:
        return ds

    rangebin = ds.coords["rangebin"].data

    return (
        ds
        .assign_coords(rangebin=np.arange(len(rangebin)) // n_bins)
        .groupby("rangebin")
        .sum()
        .assign_coords(
            rangebin=lambda x: [rangebin[i * n_bins:(i + 1) * n_bins].mean() for i in range(len(x.rangebin))])
    )


def binshift(ds: xr.Dataset, dead_bin):
    """
    Adjusts inversion bins to account for dead time or other artifacts.

    Args:
        ds (xr.Dataset): LIDAR dataset to be adjusted.
        dead_bin (int): Number of bins to shift.

    Returns:
        xr.Dataset: LIDAR dataset with adjusted bin positions.
    """
    def displace(data_: np.array, bins: int):
        data_ = data_.copy()
        if bins > 0:
            data_[:-bins] = data_[bins:]
        elif bins == 0:
            return data_
        else:
            data_ = np.insert(data_, 0, [0] * (-bins))[:bins]

        return data_

    ds = ds.copy()

    if "channel" not in ds.coords:
        ds.phy.data = displace(ds.phy.data, dead_bin)
        return ds
    dims = None
    new_phys = []
    for channel, dbin in zip(ds.channel.data, dead_bin):
        data = ds.sel(channel=channel).phy.data
        if "time" in ds.coords:
            data_times = []
            for time in ds.time.data:
                data_time = ds.sel(channel=channel, time=time).phy.data
                data_times.append(displace(data_time, dbin))
            new_phys.append(data_times)
            dims = ["channel", "time", "rangebin"]
        else:
            new_phys.append(displace(data, dbin))
            dims = ["channel", "rangebin"]

    ds = ds.assign(phy=xr.DataArray(new_phys, dims=dims))

    return ds


def get_uncertainty(signal: np.array, nshoots: np.array, background: np.array):
    """
    Computes the uncertainty for LIDAR observations.

    Args:
        signal (np.array): Observed LIDAR signal.
        nshoots (np.array): Number of laser shots.
        background (np.array): Background signal values.

    Returns:
        np.array: Uncertainty values corresponding to the input LIDAR signal.
    """
    if len(signal.shape) == 1:
        t = nshoots / 20e6
        n = t * signal * 1e6
        n_bg = t * background * 1e6

    else:
        t = (nshoots / 20e6)[:, np.newaxis]
        n = t * signal * 1e6
        n_bg = t * background.reshape(-1, 1) * 1e6

    sigma_n = ((n + n_bg) ** 0.5)
    sigma_p = sigma_n * 1e-6 / t

    return sigma_p


def dead_time_correction(lidar_data: xr.Dataset, dead_time: float):
    """
    Applies a dead-time correction to the provided LIDAR dataset.

    Args:
        lidar_data (xr.Dataset): LIDAR dataset to be corrected.
        dead_time (float): Dead-time value for correction.

    Returns:
        xr.Dataset: Corrected LIDAR dataset.
    """
    if "channel" in lidar_data.dims:
        dead_times = np.array([
            dead_time * wavelength.endswith("1") for wavelength in lidar_data.coords["channel"].data
        ]).reshape(-1, 1)
    else:
        dead_times = dead_time
    try:
        new_signals = lidar_data.phy / (1 - dead_times * lidar_data.phy)
    except:
        new_signals = lidar_data.phy / (1 - dead_times.reshape(-1, 1, 1) * lidar_data.phy)

    lidar_data.phy.data = new_signals.data

    return lidar_data


class FindFitRegion:
    min_porc = 0.7
    step = 200

    def __init__(self, signal, sigma, rangebin, molecular_data: xr.Dataset, z_ref):
        self.signal = signal
        self.sigma = sigma
        self.rangebin = rangebin
        self.z_ref = z_ref
        self.model = (molecular_data.beta.data
                      * np.exp(-2 * cumtrapz(rangebin, molecular_data.alpha.data, initial=0)) / rangebin ** 2)

    @staticmethod
    def _chi2_reduced(y_data, y_model, y_err, n_params):
        chi2 = np.sum((y_data - y_model) ** 2 / y_err ** 2)
        dof = len(y_data) - n_params
        return chi2 / dof

    @staticmethod
    def _fitter(x_data, y_data):
        if any(y_data <= 0):
            x_mean = np.mean(x_data)
            x_data = x_data / x_mean

            y_mean = np.mean(y_data)
            y_data = y_data / y_mean
            reg = np.polyfit(x_data, y_data, 1)
            return True, [reg[0] * y_mean / x_mean, reg[1] * x_mean]
        else:
            reg = np.polyfit(np.log(x_data), np.log(y_data), 1)
            return False, reg

    @staticmethod
    def _choose_model(reg, x_data, model_flag):
        """
        Evaluate the model for given regression coefficients and x_data.

        Parameters:
        - reg (array-like): Regression coefficients.
        - x_data (array-like): Independent variable inversion.
        - model_flag (bool): Model type flag.

        Returns:
        - array-like: Model output.
        """
        if model_flag:
            return reg[0] * x_data + reg[1]
        else:
            return np.exp(reg[0] * np.log(x_data) + reg[1])

    def _find_best_region(self, regions):
        """
        Finds the best region with the minimum reduced chi-squared value.

        Parameters:
        - x_data, y_data, y_err (array-like): Data arrays.
        - regions (list): List of candidate regions.

        Returns:
        - list: Region with the minimum reduced chi-squared value.
        """

        min_chi2_reduced = 999999999
        min_region = -4

        chis2 = []
        regs = []
        flags = []
        for i, region in enumerate(regions):
            x_fit = self.model[region[0]:region[1]]
            y_fit = savgol_filter(self.signal, 11, 2)
            j = 0
            while any(y_fit[region[0]:region[1]] <= 0) & (j <= 5):
                y_fit = savgol_filter(y_fit, 11, 2)
                j += 1
            y_fit = y_fit[region[0]:region[1]]
            model_flag, reg = self._fitter(x_fit, y_fit)
            chi2_reduced_current = self._chi2_reduced(self.signal[region[0]:region[1]],
                                                      self._choose_model(reg, self.model[region[0]:region[1]],
                                                                         model_flag),
                                                      self.sigma[region[0]:region[1]],
                                                      len(reg))

            chis2.append(chi2_reduced_current)
            regs.append(reg)
            flags.append(model_flag)

            if chi2_reduced_current <= min_chi2_reduced * self.min_porc:
                min_chi2_reduced = chi2_reduced_current
                min_region = i
        print(chis2)
        return regions[min_region]

    def set_min_chi2_reduction(self, min_porc):
        self.min_porc = min_porc
        return self

    def set_reference_step(self, step):
        self.step = step
        return self

    def fit(self):
        references = [[alt_bot, self.z_ref[1]]
                      for alt_bot in range(int(self.z_ref[0]), int(self.z_ref[1] - self.step * 2), self.step)]
        print(references)
        references = [z_finder(self.rangebin, ref_) for ref_ in references]
        ref = self._find_best_region(references)

        return ref



import numpy as np
from scipy.interpolate import interp1d
import datetime
from lidarpy.utils.functions import smooth, smooth_diego_fast, z_finder


class CloudFinder:
    """
    CloudFinder identifies cloud layers in LIDAR signal inversion.

    Analyzes LIDAR signals to detect cloud layers based on specific conditions and statistical measures.

    Attributes:
        _Z_MAX (int): Maximum altitude in meters.
        _original_data (np.array): Original LIDAR signal inversion.
        ref_min (int): Minimum reference altitude index.
        ref_max (int): Maximum reference altitude index.
        z (np.array): Altitude values derived from LIDAR inversion.
        signal (np.array): LIDAR signal inversion subset for reference altitudes.
        sigma (np.array): Standard deviation of the LIDAR signal inversion within reference altitudes.
        window (int): Window size for smoothing operations.
        datetime (float): Julian date representation.
    """
    _Z_MAX = 25_000

    def __init__(self, rangebin: np.array, signal: np.array, sigma: np.array, z_min: float, window: int,
                 time: datetime):
        self._original_data = signal.copy()
        self.ref_min, self.ref_max = z_finder(rangebin, z_min), z_finder(rangebin, self._Z_MAX)
        self.signal = signal[self.ref_min:self.ref_max]
        self.z = rangebin[self.ref_min:self.ref_max]
        self.sigma = sigma[self.ref_min:self.ref_max]
        self._sigma_original = sigma.copy()
        self.window = window
        self.datetime = time

    def _rcs_with_smooth(self):
        signal_w_smooth = smooth(self.signal, self.window)[::self.window]

        z_aux = self.z[::self.window]
        rcs_aux = signal_w_smooth * z_aux ** 2

        if z_aux[-1] != self.z[-1]:
            z_aux = np.append(z_aux, self.z[-1])
            rcs_aux = np.append(rcs_aux, signal_w_smooth[-1] * self.z[-1] ** 2)

        f_rcs = interp1d(z_aux, rcs_aux)
        rcs_smooth = f_rcs(self.z)
        rcs_smooth[-1] = 0
        rcs_smooth[self.z > 5000] = smooth(rcs_smooth[self.z > 5000], 1)

        rcs_smooth_2 = smooth(rcs_smooth, 69)
        rcs_smooth_3 = rcs_smooth_2.copy()

        return rcs_smooth, rcs_smooth_2, rcs_smooth_3

    def _sigma_rcs(self):
        sigma2 = self._sigma_original.copy()
        aux = np.unique(sigma2)
        aux.sort()
        if len(aux) > 1:
            sigma2[sigma2 <= aux[0]] = aux[1]

        sigma_rcs_smooth = ((smooth((sigma2[self.ref_min:self.ref_max] / self.window) ** 2, self.window)
                             * self.window) ** 0.5) * self.z ** 2

        for alt, coef in zip([7000, 5000, 3000], [1, 3, 1]):
            ref = z_finder(self.z, alt)
            sigma_rcs_smooth[:ref + 1] = coef * sigma_rcs_smooth[ref + 1] + sigma_rcs_smooth[:ref + 1]

        sigma_rcs_smooth_2 = sigma_rcs_smooth.copy()

        return sigma_rcs_smooth, sigma_rcs_smooth_2

    def _get_signal_without_cloud(self, rcs_smooth: np.array, sigma_rcs_smooth_2: np.array, rcs_smooth_sm_2: np.array,
                                  rcs_smooth_sm_3: np.array):
        rcs_smooth_exc = rcs_smooth_sm_2.copy()

        p_test = None
        asfend = 26
        for asf in range(1, asfend):
            rcs_smooth_aux = rcs_smooth.copy()
            npp = 499
            if (asf <= 15) & (asf > 1):
                test_z_rcs_smooth_sm = (rcs_smooth_aux - rcs_smooth_sm_2) / sigma_rcs_smooth_2
                mask_aux = test_z_rcs_smooth_sm > 1.5
            else:
                test_z_rcs_smooth_sm = np.abs(rcs_smooth_aux - rcs_smooth_sm_2) / sigma_rcs_smooth_2
                mask_aux = test_z_rcs_smooth_sm > 0.2

            rcs_smooth_aux[mask_aux] = rcs_smooth_sm_2[mask_aux]

            rcs_smooth_sm_2 = smooth(rcs_smooth_aux, npp) if asf != (asfend - 1) else smooth(rcs_smooth_aux, 69)

            m_aux = rcs_smooth_sm_2 > rcs_smooth_sm_3 + 0.5 * sigma_rcs_smooth_2
            rcs_smooth_sm_2[m_aux] = rcs_smooth_sm_3[m_aux]

            rcs_smooth_exc = rcs_smooth_sm_2.copy()

            if asf < 3:
                rcs_smooth_exc = rcs_smooth_exc.copy()
            else:
                r_ts = self.z > 8500
                rcs_smooth_exc[r_ts] = rcs_smooth_exc[r_ts]
            if asf == 2:
                p_test = rcs_smooth_exc.copy()
            if asf == asfend - 2:
                r_ts = self.z < 10_000
                rcs_smooth_exc[r_ts] = p_test[r_ts]
                rcs_smooth_sm_2[r_ts] = p_test[r_ts]

        return rcs_smooth_exc

    def _cloud_finder(self, rcs: np.array, sigma_rcs: np.array, rcs_smooth: np.array, rcs_smooth_exc: np.array,
                      sigma_rcs_smooth: np.array) -> tuple:
        snr_exc = rcs_smooth_exc / sigma_rcs
        ind_base, ind_top = [], []
        n1 = 2
        hour = self.datetime.hour

        k9101112 = z_finder(self.z, [9000, 10_000, 11_000, 12_000])

        pa2, pd2 = 1, self.window
        rn = pa2 + pd2 + 1

        tz_cond2 = (smooth_diego_fast(rcs - rcs_smooth_exc, pa2, pd2)
                    / ((smooth_diego_fast((sigma_rcs / rn) ** 2, pa2, pd2) * rn) ** .5))

        r = (self.z < 5000) | (self.z > 22_000)
        tz_cond2[r] = 0

        k = 2
        cont = 0

        while k <= z_finder(self.z, 20_000):
            k += 1
            cond1 = ((rcs_smooth[k + 0 * self.window] > (rcs_smooth_exc[k + 0 * self.window]
                                                         + n1 * sigma_rcs_smooth[k + 0 * self.window]))
                     & (rcs_smooth[k + 1 * self.window] > (rcs_smooth_exc[k + 1 * self.window]
                                                           + n1 * sigma_rcs_smooth[k + 1 * self.window]))
                     & (rcs_smooth[k + 2 * self.window] > (rcs_smooth_exc[k + 2 * self.window]
                                                           + n1 * sigma_rcs_smooth[k + 2 * self.window])))
            cond2 = (tz_cond2[k] > 4) & (self.z[k] > 10_000) & (snr_exc[k] > 0.01)
            cond = cond1 | cond2
            if cond:
                if (hour <= 6) | (hour >= 18):
                    if (self.z[k] > 10_000) & (sum(snr_exc[k9101112] < 0.1) > 0):
                        continue
                cont += 1
                ind_base.append(k)
                if ind_base[-1] <= 0:
                    ind_base = 1

                for kk in range(k + self.window, len(self.z)):
                    if (rcs_smooth[kk] < rcs_smooth_exc[k]) & (rcs_smooth[kk] < rcs_smooth_exc[kk]):
                        ind_top.append(kk)
                        k = kk + 1
                        break

        return ind_base, ind_top

    def _distance_between_clouds_correction(self):
        pass
        if len(self.z_base) > 1:
            ind_clear_zt = np.array(np.where((self.z_base[1:] - self.z_top[:-1]) < 500)[0])
            ind_clear_zb = ind_clear_zt + 1
            self.z_top = np.delete(self.z_top, ind_clear_zt)
            self.z_base = np.delete(self.z_base, ind_clear_zb)

    def _comp(self, rcs_smooth_exc, ind_base, ind_top):
        if (ind_base == []) | (ind_top == []):
            self.z_base, self.z_top, self.z_max_capa, self.nfz_base, self.nfz_top, self.nfz_max_capa = [[np.nan]] * 6
            return self.z_base, self.z_top, self.z_max_capa, self.nfz_base, self.nfz_top, self.nfz_max_capa

        z_base = self.z[ind_base]
        z_top = self.z[ind_top]

        self.z_base, self.z_top = np.array(z_base), np.array(z_top)

        self._distance_between_clouds_correction()

        new_ind_base = z_finder(self.z, self.z_base)
        new_ind_top = z_finder(self.z, self.z_top)

        nfz_base = rcs_smooth_exc[new_ind_base]
        nfz_top = rcs_smooth_exc[new_ind_top]
        nfz_max_capa = []
        z_max_capa = []
        for base, top in zip(new_ind_base, new_ind_top):
            max_signal = max(self.signal[base:top])
            bools = self.signal[base:top] == max_signal
            z_max = self.z[base:top][np.where(bools)[0][0]]
            nfz_max_capa.append(max_signal * z_max ** 2)
            z_max_capa.append(z_max)

        self.z_max_capa, self.nfz_base, self.nfz_top, self.nfz_max_capa = (np.array(z_max_capa),
                                                                           np.array(nfz_base),
                                                                           np.array(nfz_top),
                                                                           np.array(nfz_max_capa))

        return self.z_base, self.z_top, self.z_max_capa, self.nfz_base, self.nfz_top, self.nfz_max_capa

    def fit(self):
        rcs_smooth, rcs_smooth_2, rcs_smooth_3 = self._rcs_with_smooth()

        sigma_rcs_smooth, sigma_rcs_smooth_2 = self._sigma_rcs()

        rcs_smooth_exc = self._get_signal_without_cloud(rcs_smooth, sigma_rcs_smooth_2, rcs_smooth_2, rcs_smooth_3)

        ind_base, ind_top = self._cloud_finder(self.signal * self.z ** 2,
                                               self.sigma * self.z ** 2,
                                               rcs_smooth,
                                               rcs_smooth_exc,
                                               sigma_rcs_smooth)

        return self._comp(rcs_smooth_exc, ind_base, ind_top)


def cloud_mask_correction(z_bases, z_tops, z_max_capas, nfz_bases, nfz_tops, nfz_max_capas):
    z_bases_, z_tops_, z_max_capas_, nfz_bases_, nfz_tops_, nfz_max_capas_ = (z_bases.copy(), z_tops.copy(),
                                                                              z_max_capas.copy(), nfz_bases.copy(),
                                                                              nfz_tops.copy(), nfz_max_capas.copy())
    """
    Corrects cloud masks based on proximity criteria between different layers of cloud inversion.

    Parameters:
    - z_bases (list of lists of floats): Base heights of the clouds.
    - z_tops (list of lists of floats): Top heights of the clouds.
    - z_max_capas (list of lists of floats): Maximum heights within the cloud layers.
    - nfz_bases (list of lists of floats): Base heights of the no-fly zones.
    - nfz_tops (list of lists of floats): Top heights of the no-fly zones.
    - nfz_max_capas (list of lists of floats): Maximum heights within the no-fly zone layers.

    Returns:
    - tuple of lists: Corrected versions of z_bases, z_tops, z_max_capas, nfz_bases, nfz_tops, and nfz_max_capas.

    Notes:
    The function works by iterating over the cloud inversion and checking for nearby layers within specified thresholds.
    Layers that match the criteria are considered neighbors and are used in the correction process. The function also 
    considers a temporal component in the checking process, allowing for comparison of layers from nearby time steps.

    The function makes use of several hard-coded parameters:
    - dt: Temporal proximity for the comparison.
    - dz1 and dz2: Vertical proximity thresholds based on the thickness of the layer.
    - dthick: Threshold for determining which dz value to use.
    """
    dt = 1
    dz1 = 500
    dz2 = 300
    dthick = 1e3
    for u, (bases, tops) in enumerate(zip(z_bases, z_tops)):
        if np.isnan(bases)[0]:
            continue
        neib = np.zeros(len(bases), dtype=bool)
        for i, (base, top) in enumerate(zip(bases, tops)):
            uaux = list(range(u - dt, u + dt + 1))
            for uj in uaux:
                if (uj < 0) | (uj == len(z_bases)) | (uj == u):
                    continue
                elif (np.isnan(z_bases[uj]).sum() > 0) | (np.isnan(z_tops[uj]).sum() > 0):
                    continue
                elif top - base > dthick:
                    dz = dz1
                    cond1 = sum(abs(base - z_bases[uj]) < dz) > 0
                    cond2 = sum(abs(base - z_tops[uj]) < dz) > 0
                    cond3 = sum(abs(top - z_bases[uj]) < dz) > 0
                    cond4 = sum(abs(base - z_tops[uj]) < dz) > 0
                    if cond1 | cond2 | cond3 | cond4:
                        neib[i] = True
                else:
                    dz = dz2
                    cond1 = sum(abs(base - z_bases[uj]) < dz) > 0
                    cond2 = sum(abs(base - z_tops[uj]) < dz) > 0
                    cond3 = sum(abs(top - z_bases[uj]) < dz) > 0
                    cond4 = sum(abs(base - z_tops[uj]) < dz) > 0
                    if cond1 | cond2 | cond3 | cond4:
                        neib[i] = True

        z_bases_[u] = ([np.nan] if z_bases[u][neib].size == 0 else z_bases[u][neib])
        z_tops_[u] = ([np.nan] if z_tops[u][neib].size == 0 else z_tops[u][neib])
        z_max_capas_[u] = ([np.nan] if z_max_capas[u][neib].size == 0 else z_max_capas[u][neib])
        nfz_bases_[u] = ([np.nan] if nfz_bases[u][neib].size == 0 else nfz_bases[u][neib])
        nfz_tops_[u] = ([np.nan] if nfz_tops[u][neib].size == 0 else nfz_tops[u][neib])
        nfz_max_capas_[u] = ([np.nan] if nfz_max_capas[u][neib].size == 0 else nfz_max_capas[u][neib])

    return z_bases_, z_tops_, z_max_capas_, nfz_bases_, nfz_tops_, nfz_max_capas_



import numpy as np
from scipy.signal import savgol_filter
from sklearn.linear_model import LinearRegression


def dif(x: np.array, y: np.array, window: int, weights: np.array = None):
    def fit(init, final) -> None:
        y_fit = y[init: final].reshape(-1, 1)
        x_fit = x[init: final].reshape(-1, 1)
        weight_fit = None if weights is None else weights[init: final]

        linear_regession = LinearRegression().fit(x_fit, y_fit, sample_weight=weight_fit)

        return linear_regession.coef_[0][0]

    if window % 2 == 0:
        raise ValueError("window must be odd.")

    win = window // 2
    diff_y = []
    for i in range(win, len(y) - win - 10 - 1):
        diff_y.append(fit(i - win, i + win + 1))

    for i in range(window // 2):
        diff_y.insert(0, diff_y[0])

    while len(diff_y) != len(y):
        diff_y += [diff_y[-1]]

    return np.array(diff_y)


def diff_linear_regression(diff_window: int):
    def diff(raman):
        weights = (None if raman.inelastic_uncertainty is None
                   else 1 / (raman.inelastic_uncertainty * raman.rangebin ** 2) ** 2)

        ranged_corrected_signal = raman.inelastic_signal * raman.rangebin ** 2

        dif_ranged_corrected_signal = dif(raman.rangebin, ranged_corrected_signal, diff_window, weights)
        dif_num_density = dif(raman.rangebin, raman.raman_scatterer_numerical_density, diff_window)
        return ((dif_num_density / raman.raman_scatterer_numerical_density)
                - (dif_ranged_corrected_signal / ranged_corrected_signal))

    return diff


def get_savgol_filter(window_length, polyorder):
    def diff(raman):
        ranged_corrected_signal = raman.inelastic_signal * raman.rangebin ** 2
        num_density = savgol_filter(raman.raman_scatterer_numerical_density, window_length, polyorder)
        ranged_corrected_signal = savgol_filter(ranged_corrected_signal, window_length, polyorder)

        dif_ranged_corrected_signal = np.gradient(ranged_corrected_signal) / np.gradient(raman.rangebin)
        dif_num_density = np.gradient(num_density) / np.gradient(raman.rangebin)
        return (dif_num_density / num_density) - (dif_ranged_corrected_signal / ranged_corrected_signal)

    return diff


def beta_smooth(raman):
    return raman.get_beta()["elastic_aer"]



import re
import numpy as np
from datetime import datetime
import xarray as xr


class GetData:
    """
    Handles fetching and processing of LIDAR inversion from binary files.

    This class provides functionality to read binary files containing LIDAR inversion, extracting
    headers and inversion content. After processing, the inversion can be converted into xarray format for efficient analysis.

    Attributes:
        directory (str): Directory containing the LIDAR inversion files.
        files_name (list[str]): List of filenames to be processed.
    """
    files_w_error = []

    def __init__(self, directory: str, files_name: list) -> None:
        self.directory = directory
        self.files_name = files_name.copy()
        self.files_name.sort()

    @staticmethod
    def profile_read(f_name: str) -> tuple:
        """
        Extracts header and inversion content from a LIDAR binary file.

        The binary file's header contains metadata about the LIDAR observation, including
        location, date, altitude, and more. This method retrieves that information along with
        the observation inversion.

        Args:
            f_name (str): Path to the binary LIDAR inversion file.

        Returns:
            tuple:
                - dict: Header information.
                - np.array: LIDAR observation in physical units.
                - np.array: Raw LIDAR observation.
        """
        with open(f_name, 'r', encoding='utf8', errors='ignore') as fp:
            #  Linha 1
            regexp = re.compile('([\w]{9}.[\d]{3})')  # filename

            line = regexp.search(fp.readline())

            head = {'file': line.group(1)}

            #  Linha 2
            regexp = re.compile(' ([\w]*) '  # site
                                '([\d]{2}/[\d]{2}/[\d]{4}) '  # datei
                                '([\d]{2}:[\d]{2}:[\d]{2}) '  # houri
                                '([\d]{2}/[\d]{2}/[\d]{4}) '  # datef
                                '([\d]{2}:[\d]{2}:[\d]{2}) '  # hourf
                                '([\d]{4}) '  # alt
                                '(-?[\d]{3}\.\d) '  # lon
                                '(-?[\d]{3}\.\d) '  # lat
                                '(-?[\d]{1,2}) '  # zen
                                '[\d]{2} '  # ---- empty
                                '([\d]{2}\.\d) '  # T0
                                '([\d]{4}\.\d)')  # P0

            line = regexp.search(fp.readline())

            head['site'] = line.group(1)
            head['datei'] = line.group(2)
            head['houri'] = line.group(3)
            head['datef'] = line.group(4)
            head['hourf'] = line.group(5)

            def date_num(d):
                return 366 + d.toordinal() + (d - datetime.fromordinal(d.toordinal())).total_seconds() / (24 * 60 * 60)

            jdi = head['datei'] + ' ' + head['houri']
            jdi_strip = datetime.strptime(jdi, '%d/%m/%Y %H:%M:%S')

            jdf = head['datef'] + ' ' + head['hourf']
            jdf_strip = datetime.strptime(jdf, '%d/%m/%Y %H:%M:%S')

            head['jdi'] = date_num(jdi_strip)
            head['jdf'] = date_num(jdf_strip)

            head['alt'] = int(line.group(6))
            head['lon'] = float(line.group(7))
            head['lat'] = float(line.group(8))
            head['zen'] = float(line.group(9))
            head['T0'] = float(line.group(10))
            head['P0'] = float(line.group(11))

            #  Linha 3
            regexp = re.compile('([\d]{7}) '  # nshoots    
                                '([\d]{4}) '  # nhz
                                '([\d]{7}) '  # nshoots2
                                '([\d]{4}) '  # nhz2
                                '([\d]{2}) ')  # nch

            line = regexp.search(fp.readline())

            head['nshoots'] = int(line.group(1))
            head['nhz'] = int(line.group(2))
            head['nshoots2'] = int(line.group(3))
            head['nhz2'] = int(line.group(4))
            head['nch'] = int(line.group(5))

            #  Canais
            head['ch'] = {}
            nch = head['nch']  # Número de canais

            regexp = re.compile('(\d) '  # active
                                '(\d) '  # photons
                                '(\d) '  # elastic
                                '([\d]{5}) '  # ndata
                                '\d '  # ----
                                '([\d]{4}) '  # pmtv
                                '(\d\.[\d]{2}) '  # binw
                                '([\d]{5})\.'  # wlen
                                '([osl]) '  # pol
                                '[0 ]{10} '  # ----
                                '([\d]{2}) '  # bits
                                '([\d]{6}) '  # nshoots
                                '(\d\.[\d]{3,4}) '  # discr
                                '([\w]{3})')  # tr

            channels = ''.join([next(fp) for _ in range(nch)])  # Aqui eu imprimo todos os canais

            lines = np.array(regexp.findall(channels))

            head['ch']['active'] = lines[:, 0].astype(int)
            head['ch']['photons'] = lines[:, 1].astype(int)
            head['ch']['elastic'] = lines[:, 2].astype(int)
            head['ch']['ndata'] = lines[:, 3].astype(int)
            head['ch']['pmtv'] = lines[:, 4].astype(int)
            head['ch']['binw'] = lines[:, 5].astype(float)
            head['ch']['wlen'] = lines[:, 6].astype(int)
            head['ch']['pol'] = lines[:, 7]
            head['ch']['bits'] = lines[:, 8].astype(int)
            head['ch']['nshoots'] = lines[:, 9].astype(int)
            head['ch']['discr'] = lines[:, 10].astype(float)
            head['ch']['tr'] = lines[:, 11]

            # Criei os arrays phy e raw antes, pois no matlab elas são criadas enquanto declaradas

            max_linhas = max(head['ch']['ndata'])  # A solucao que encontrei aqui foi achar o max de
            # linhas possivel que phy e raw podem ter para declarar antes

            phy = np.zeros((max_linhas, nch))
            raw = np.zeros((max_linhas, nch))

            # conversion factor from raw to physical units
            for ch in range(nch):
                nz = head['ch']['ndata'][ch]
                _ = np.fromfile(fp, np.byte, 2)
                tmpraw = np.fromfile(fp, np.int32, nz)

                if head['ch']['photons'][ch] == 0:
                    d_scale = head['ch']['nshoots'][ch] * (2 ** head['ch']['bits'][ch]) / (head['ch']['discr'][ch]
                                                                                           * 1e3)
                else:
                    d_scale = head['ch']['nshoots'][ch] / 20

                tmpphy = tmpraw / d_scale

                # copy to final destination
                phy[:nz, ch] = tmpphy[:nz]
                raw[:nz, ch] = tmpraw[:nz]

        return head, phy.T, raw.T

    def get_xarray(self):
        """
        Transforms LIDAR inversion from binary files into xarray format.

        Processes each LIDAR inversion file, extracts headers and inversion content, and consolidates
        the extracted inversion into an xarray format. This structured format aids further manipulation and analysis.

        Returns:
            xr.Dataset: LIDAR inversion consolidated in xarray format.
        """
        (times, datei, houri, datef, hourf, jdi, jdf, pressures_0, temperatures_0, phys, raws, nshoots, zen,
         wavelengths_, actives, photons, elastic, ndata, pmtv, binw, pol, bits, tr, discr) = ([], [], [], [], [], [],
                                                                                              [], [], [], [], [], [],
                                                                                              [], [], [], [], [], [],
                                                                                              [], [], [], [], [], [])
        count = 0
        length = None
        first_head = None
        for file in self.files_name:
            try:
                head, phy, raw = self.profile_read(f"{self.directory}/{file}")
            except:
                count += 1
                self.files_w_error.append(file)
                if count == len(self.files_name):
                    print(f"problemas={count} de {len(self.files_name)}")
                    return None
                continue

            if first_head is None:
                first_head = head.copy()
                for key in ["file", "datei", "houri", "datef", "hourf", "jdi", "jdf", "T0", "P0", "ch", "nshoots",
                            "nshoots2", "zen", "site"]:
                    first_head.pop(key)
                length = len(phy[0, :])
            else:
                new_head = head.copy()
                for key in ["file", "datei", "houri", "datef", "hourf", "jdi", "jdf", "T0", "P0", "ch", "nshoots",
                            "nshoots2", "zen", "site"]:
                    new_head.pop(key)
                bool1 = first_head == new_head
                if not bool1:
                    raise Exception("All headers in the directory must be the same.")
            phys.append(phy[:, :length])
            raws.append(raw[:, :length])
            times.append((head["jdi"] + head["jdf"]) / 2)

            vecs1 = [datei, houri, datef, hourf, jdi, jdf, pressures_0, temperatures_0, zen]
            keys1 = ["datei", "houri", "datef", "hourf", "jdi", "jdf", "P0", "T0", "zen"]
            for vec, key in zip(vecs1, keys1):
                vec.append(head[key])

            vecs2 = [nshoots, wavelengths_, actives, photons, elastic, ndata, pmtv, binw, pol, bits, tr, discr]
            keys2 = ["nshoots", "wlen", "active", "photons", "elastic", "ndata", "pmtv", "binw", "pol", "bits", "tr",
                     "discr"]
            for vec, key in zip(vecs2, keys2):
                vec.append(head["ch"][key])

        wavelengths = [f"{wavelength}_{photon}"
                       for wavelength, photon
                       in zip(head["ch"]["wlen"], head["ch"]["photons"])]

        rangebin = np.arange(1, len(phys[0][0]) + 1) * head["ch"]["binw"][0]

        das = {"phy": xr.DataArray(phys, coords=[times, wavelengths, rangebin], dims=["time", "channel", "rangebin"]),
               "raw": xr.DataArray(raws, coords=[times, wavelengths, rangebin], dims=["time", "channel", "rangebin"])}

        vars1 = [datei, houri, datef, hourf, jdi, jdf, pressures_0, temperatures_0, zen]
        names1 = ["datei", "houri", "datef", "hourf", "jdi", "jdf", "pressure0", "temperature0", "zen"]
        for name1, var1 in zip(names1, vars1):
            das[name1] = xr.DataArray(var1, coords=[times], dims="time")

        vars2 = [nshoots, wavelengths_, actives, photons, elastic, ndata, pmtv, binw, pol, bits, tr, discr]
        names2 = ["nshoots", "wlen", "active", "photon", "elastic", "ndata", "pmtv", "binw", "pol", "bits",
                  "tr", "discr"]
        for name2, var2 in zip(names2, vars2):
            das[name2] = xr.DataArray(var2, coords=[times, wavelengths], dims=["time", "channel"])
        ds = xr.Dataset(das)

        first_head["site"] = head["site"]

        ds.attrs = first_head

        ds = ds.assign(rcs=lambda x: ds.phy * ds.coords["rangebin"].data ** 2)
        return ds



import numpy as np
import xarray as xr
from scipy.integrate import cumtrapz, trapz
from scipy.optimize import curve_fit
from lidarpy.utils.functions import z_finder


def calib_strategy1(signal: np.array,
                    model: np.array,
                    reference: np.array):
    """
    Calibrate a signal against a model using a linear strategy.

    Parameters:
    ----------
    signal : np.array
        Input signal to be calibrated.
    model : np.array
        Model against which the signal is calibrated.
    reference : np.array
        Indices considered for calibration.

    Returns:
    -------
    tuple
        A tuple containing the calibrated signal, model, and the coefficients of the linear fit.
    """
    reference -= 1

    if reference[-1] >= len(signal):
        reference = np.arange(reference[0], len(signal))

    mean_signal = np.mean(signal[reference])
    signal = signal / mean_signal

    mean_model = np.mean(model[reference])
    model = model / mean_model

    coef_slope_linear, *_ = curve_fit(f=lambda x, a, b: a * x + b, xdata=model[reference], ydata=signal[reference])

    signal = (signal - coef_slope_linear[1]) / coef_slope_linear[0]

    coef_slope_linear = [coef_slope_linear[0] * mean_signal / mean_model, coef_slope_linear[1] * mean_signal]

    return signal, model, coef_slope_linear


def calib_strategy2(signal: np.array,
                    model: np.array,
                    reference: np.array):
    """
    Calibrate a signal against a model using a simplified linear strategy.

    Parameters:
    ----------
    signal : np.array
        Input signal to be calibrated.
    model : np.array
        Model against which the signal is calibrated.
    reference : np.array
        Indices considered for calibration.

    Returns:
    -------
    tuple
        A tuple containing the calibrated signal, model, and the coefficient of the fit.
    """
    mean_signal = np.mean(signal[reference])
    signal = signal / mean_signal

    mean_model = np.mean(model[reference])
    model = model / mean_model

    coef_slope, *_ = curve_fit(f=lambda x, a: a * x, xdata=model[reference], ydata=signal[reference])

    signal = signal / coef_slope[0]

    coef_slope = [coef_slope[0] * mean_signal / mean_model]

    return signal, model, coef_slope


class Klett:
    """
    Implements the Klett inversion algorithm for lidar inversion.

    Attributes:
    ----------
    z : array
        Height profile in meters.
    pr : array
        Lidar profile.
    ref : array
        Reference height(s) in meters.
    lambda_ : float
        Lidar wavelength in nanometers.
    lidar_ratio : float
        Extinction to backscatter ratio in sr.
    p_air : array
        Atmospheric pressure profile in Pa.
    t_air : array
        Atmospheric temperature profile in K.
    co2ppmv : float
        CO2 concentration in ppmv.

    Methods:
    -------
    get_beta() -> dict:
        Returns aerosol backscatter coefficients.
    get_alpha() -> dict:
        Returns aerosol extinction coefficients.
    get_lidar_ratio() -> dict:
        Returns the lidar ratios.
    set_lidar_ratio(lidar_ratio: float):
        Sets a new lidar ratio.
    get_model_mol() -> np.array:
        Returns the molecular model for the lidar signal.
    fit() -> tuple:
        Fits the lidar inversion using the Klett inversion algorithm.

    Authors:
    - Pablo Ristori (pablo.ristori@gmail.com) CEILAP, UNIDEF (CITEDEF-CONICET), Argentina
    - Lidia Otero (lidia1116@gmail.com) CEILAP, UNIDEF (CITEDEF-CONICET), Argentina
    """
    _alpha = dict()
    _beta = dict()
    _lr = dict()
    tau = None
    tau_std = None
    mc_iter = None
    fit_parameters = None
    _calib_strategies = {True: calib_strategy1, False: calib_strategy2}

    def __init__(self, rangebin: np.array, signal: np.array, molecular_data: xr.Dataset, lidar_ratio: float,
                 molecular_reference_region: list, correct_noise: bool = True):
        self.signal = signal.copy()
        self.rangebin = rangebin.copy()
        self.ref = molecular_reference_region
        self._calib_strategy = self._calib_strategies[correct_noise]
        self._alpha['mol'], self._beta['mol'], self._lr['mol'] = (molecular_data.alpha.data, molecular_data.beta.data,
                                                                  molecular_data.lidar_ratio.data)
        self._lr['aer'] = lidar_ratio

    def __str__(self):
        return f"Lidar ratio = {self._lr['aer']}"

    def get_beta(self) -> dict:
        return self._beta.copy()

    def get_alpha(self) -> dict:
        return self._alpha.copy()

    def get_lidar_ratio(self) -> dict:
        return self._lr.copy()

    def set_lidar_ratio(self, lidar_ratio: float):
        self._lr['aer'] = lidar_ratio
        return self

    def set_correction(self, correct_noise: bool):
        self._calib_strategy = self._calib_strategies[correct_noise]
        return self

    def get_model_mol(self) -> np.array:
        return (self._beta['mol'] * np.exp(-2 * cumtrapz(self._alpha['mol'], self.rangebin, initial=0))
                / self.rangebin ** 2)

    def _calibration(self, signal):
        ref = z_finder(self.rangebin, self.ref)

        if len(ref) > 1:
            signal, model, self.fit_parameters = self._calib_strategy(signal=signal.copy(),
                                                                      model=self.get_model_mol(),
                                                                      reference=np.arange(*ref))

            beta_ref = self._beta['mol'][ref[0]] * signal[ref[0]] / model[ref[0]]
        else:
            signal = signal
            beta_ref = self._beta['mol'][ref[0]]

        return beta_ref, signal, ref[0]

    def fit(self):
        beta_ref, signal, ref0 = self._calibration(self.signal)

        corrected_signal = signal * self.rangebin ** 2

        spp = corrected_signal * np.exp(- 2 * cumtrapz(x=self.rangebin,
                                                       y=(self._lr['aer'] - self._lr['mol']) * self._beta['mol'],
                                                       initial=0))

        sppr = spp / spp[ref0]

        self._beta['tot'] = sppr / (
                1 / beta_ref - (cumtrapz(x=self.rangebin, y=2 * self._lr['aer'] * sppr, initial=0)
                                - trapz(x=self.rangebin[:ref0],
                                        y=2 * (self._lr['aer'] * np.ones_like(sppr))[:ref0] * sppr[:ref0])))

        self._beta['aer'] = self._beta['tot'] - self._beta['mol']

        self._alpha['aer'] = self._beta['aer'] * self._lr['aer']

        self._alpha['tot'] = self._alpha['mol'] + self._alpha['aer']

        self.tau = trapz(self._alpha["aer"], self.rangebin)

        return self._alpha["aer"].copy(), self._beta["aer"].copy(), self._lr["aer"]



import numpy as np
from scipy.interpolate import interp1d
from scipy.integrate import cumtrapz
import datetime as dt


def matlab2datetime(matlab_datenum):
    """
        Converts a MATLAB datenum into a Python datetime object.

        Parameters:
        -----------
        matlab_datenum : float
            The MATLAB datenum to be converted.

        Returns:
        --------
        datetime.datetime
            Corresponding Python datetime object.

        Note:
        -----
        MATLAB's datenum starts from year 0, whereas Python's datetime starts from year 1. This function
        adjusts for that difference.
    """
    day = dt.datetime.fromordinal(int(matlab_datenum))
    dayfrac = dt.timedelta(days=matlab_datenum%1) - dt.timedelta(days=366)
    return day + dayfrac


# def molecular_raman_model(signal: np.array, rangebin: np.array, lidar_wavelength, raman_wavelength, p_air, t_air,
# alt_ref, co2ppmv=375) -> np.array:
#     """
#     Calculate the molecular Raman model.
#
#     Parameters:
#     - lidar_data: Input lidar dataset.
#     - lidar_wavelength: Lidar operating wavelength.
#     - raman_wavelength: Raman scattering wavelength.
#     - p_air: Air pressure.
#     - t_air: Air temperature.
#     - alt_ref: Altitude reference.
#     - co2ppmv: Carbon dioxide concentration in ppmv. Default is 375.
#     - pc: Photocount mode. Default is True.
#
#     Returns:
#     - np.array: Molecular Raman model.
#     """
#     alpha_lidar_mol, *_ = AlphaBetaMolecular(p_air, t_air, lidar_wavelength, co2ppmv).get_params()
#
#     alpha_raman_mol, *_ = AlphaBetaMolecular(p_air, t_air, raman_wavelength, co2ppmv).get_params()
#
#     scatterer_numerical_density = 78.08e-2 * p_air / (1.380649e-23 * t_air)
#
#     model = (scatterer_numerical_density * np.exp(-cumtrapz(rangebin, alpha_lidar_mol + alpha_raman_mol, initial=0))
#              / rangebin ** 2)
#
#     ref = z_finder(rangebin, alt_ref)
#
#     model_fit = np.log(model[ref[0]:ref[1]])
#
#     signal_fit = np.log(signal[ref[0]:ref[1]])
#
#     nans = (np.isnan(signal_fit) == False)
#
#     reg = np.polyfit(model_fit[nans], signal_fit[nans], 1)
#
#     return np.exp(reg[0] * np.log(model) + reg[1])


def smooth(signal: np.array, window: int):
    """
    Smooth a given signal using a moving average.

    Parameters:
    - signal: Input signal array.
    - window: Window size for smoothing.

    Returns:
    - np.array: Smoothed signal.
    """
    if window % 2 == 0:
        raise Exception("Window value must be odd")
    out0 = np.convolve(signal, np.ones(window, dtype=int), 'valid') / window
    r = np.arange(1, window - 1, 2)
    start = np.cumsum(signal[:window - 1])[::2] / r
    stop = (np.cumsum(signal[:-window:-1])[::2] / r)[::-1]
    return np.concatenate((start, out0, stop))


def smooth_diego_fast(signal: np.array, p_before: int, p_after: int):
    """
    Fast smoothing method implemented by Gouveia 2018.

    Parameters:
    - signal: Input signal array.
    - p_before: Points before.
    - p_after: Points after.

    Returns:
    - np.array: Smoothed signal.
    """
    sm = p_before + p_after + 1
    y_sm = smooth(signal, sm)
    y_sm4 = np.zeros(y_sm.shape)
    y_sm4[:-sm // 2 + 1] = y_sm[sm // 2:]
    return y_sm4


def signal_smoother(rangebin: np.array, signal: np.array, window: int):
    """
    Smooth a signal.

    Parameters:
    - signal: Input signal array.
    - rangebin: Range bin array.
    - window: Window size for smoothing.

    Returns:
    - np.array: Smoothed signal.
    """
    vec_smooth = smooth(signal, window)
    vec_aux = vec_smooth[::window]
    z_aux = rangebin[::window]
    if z_aux[-1] < rangebin[-1]:
        ind = z_finder(rangebin, z_aux[-1])
        vec_aux = np.append(vec_aux, signal[ind:])
        z_aux = np.append(z_aux, rangebin[ind:])
    func = interp1d(z_aux, vec_aux)

    return func(rangebin)


def z_finder(rangebin: np.array, alts):
    """
    Find the indices of altitudes in the rangebin.

    Parameters:
    - rangebin: Range bin array.
    - alts: Desired altitudes.

    Returns:
    - Indices corresponding to the given altitudes.
    """
    def finder(z: int):
        return round((z - rangebin[0]) / (rangebin[1] - rangebin[0]))

    try:
        iter(alts)
        iterable = True
    except TypeError:
        iterable = False

    if alts is None:
        return None
    elif iterable:
        return [finder(alt) for alt in alts]
    else:
        return finder(alts)


def molecular_model(rangebin, signal, molecular_data, alt_ref) -> np.array:
    alpha_mol, beta_mol = molecular_data.alpha.data, molecular_data.beta.data

    model = beta_mol * np.exp(-2 * cumtrapz(rangebin, alpha_mol, initial=0)) / rangebin ** 2

    ref = z_finder(rangebin, alt_ref)

    reg = np.polyfit(np.log(model[ref[0]:ref[1]]),
                     np.log(signal[ref[0]:ref[1]]),
                     1)

    return np.exp(reg[0] * np.log(model) + reg[1])



import subprocess
import os
import random
import numpy as np


def multiscatter(options: str = None, wavelength: float = None, alt: float = None, rho_div: float = None,
                 rho_fov: np.ndarray = None, range_: np.ndarray = None, ext: np.ndarray = None,
                 radius: np.ndarray = None, S: np.ndarray = None, ext_air: np.ndarray = None, ssa: np.ndarray = None,
                 g: np.ndarray = None, ssa_air: np.ndarray = None, droplet_fraction: np.ndarray = None,
                 pristine_ice_fraction: np.ndarray = None, bscat_AD: np.ndarray = None,
                 bscat_air_AD: np.ndarray = None):
    """
    MULTISCATTER  Perform multiple scattering calculation for radar or lidar

     This function calculates the apparent backscatter coefficient
     accounting for multiple scattering, calling the "multiscatter" C
     code.  It can be called in a number of different ways.

     For lidar multiple scattering including both small-angle (SA) and
     wide-angle (WA) scattering:

       inversion = multiscatter(options, wavelength, alt, rho_div, rho_fov, ...
                           range, ext, radius, S, ext_air, ssa, g, ...
                           ssa_air, droplet_fraction, pristine_ice_fraction)

     where the input variables are:
       options: string containing space-separated settings for the code -
         for the default options use the empty string ''
       wavelength: instrument wavelength in m
       alt: instrument altitude in m
       rho_div: half-angle 1/e beam divergence in radians
       rho_fov: half-angle receiver field-of-view in radians (or vector of FOVs)
       range: vector of ranges at which the input inversion are located, in
          m; the first range should be the closest to the instrument
       ext: vector of extinction coefficients, in m-1
       radius: vector of equivalent-area radii, in m
       S: vector of backscatter-to-extinction ratios, in sr
       ext_air: vector of molecular extinction coefficients, in m-1
       ssa: vector of particle single-scatter albedos
       g: vector of scattering asymmetry factors
       ssa_air: vector of molecular single-scatter albedos
       droplet_fraction: vector of fractions (between 0 and 1) of the particle
          backscatter that is due to droplets - only affects small-angle
          returns
       pristine_ice_fraction: vector of fractions (between 0 and 1) of the
          particle backscatter that is due to pristine ice - only affects
          small-angle returns
       bscat_AD: adjoint input for backscatter values
       bscat_air_AD: adjoint input for the backscatter of air

     For radar multiple scattering use the same form but with options
     containing the string '-no-forward-lobe'

     If droplet_fraction and pristine_ice_fraction are omitted then they are
     assumed zero, i.e. the near-backscatter phase function is assumed
     isotropic.  If "ssa_air" is also omitted it is assumed to be 1 if
     wavelength < 1e-6 m and 0 if wavelength > 1e-6. If "ssa" and "g" are
     omitted then the WA part of the calculation is not performed. If
     "ext_air" is omitted then it is assumed to be zero.

     The "options" string contains a space-separated list of arguments
     passed directly to the executable.  Valid ones to use from matlab are:

    General options
     -single-only    Single scattering only
     -qsa-only       Don't include wide-angle multiple scattering
     -wide-only      Only wide-angle multiple scattering
     -hsrl           Output particulate and air backscatter separately
     -gaussian-receiver
                     Receiver is Gaussian rather than top-hat shaped
     -adjoint        Output the adjoint as well
     -numerical-jacobian
                     Output the Jacobian instead
     -jacobian       Approximate but faster Jacobian: only for small-angle
     -ext-only       Only calculate the Jacobian with respect to extinction

    Options for quasi-small-angle (QSA) multiple scattering calculation
     -explicit n     Use an explicit model with n orders of scattering
     -fast-qsa       Use fast O(N) QSA model
     -wide-angle-cutoff <theta>
                     Forward scattering at angles greater than <theta> radians
                     are deemed to escape, a crude way to deal with a problem
                     associated with aerosols
     -approx-exp     Appriximate the exp() function call for speed

    Options for wide-angle multiple-scattering
     -no-forward-lobe
                     Radar-like phase function behaviour: use single-scattering
                     rather than QSA
     -simple-2s-coeffts
                     Use the simple upwind Euler formulation (can be unstable
                     for high optical depth)
     -ssa-scales-forward-lobe
                     Single-scattering albedo less than unity reduces the
                     forward scattering lobe as well as wide-angle scattering
     -num-samples m  Output m samples, allowing sampling of signals appearing
                     to originate below ground

     The output is written to the structure "inversion" containing the member
     "bscat" which is the apparent backscatter coefficient. It also
     includes "range", "ext" and "radius", which are the same as the input
     values.

     If the "-adjoint" option is specified then the adjoint variables
     ext_AD, ssa_AD, g_AD and ext_bscat_ratio_AD will also be returned in
     inversion based on the input variables bscat_AD and bscat_air_AD.

     Note that this function may need to be edited to indicate where the
     "multiscatter" executable file is located.
    """
    nargin = len([var for var in locals().values() if var is not None])

    multiscatter_exec = os.path.join(os.getcwd(), "multiscatter.exe")
    in_temp = "MSCATTER_" + str(random.randint(0, 10000000)) + ".DAT"
    out_temp = "STDERR_" + str(random.randint(0, 10000000)) + ".txt"

    # Check existence of executable
    if not os.path.isfile(multiscatter_exec):
        raise Exception(f"{multiscatter_exec} not found")

    # If fewer than 9 arguments, show the help
    if nargin < 9:
        if nargin == 1 and options == "executable":
            return multiscatter_exec
        print("help multiscatter")
        return

    # Useful vectors
    z = np.zeros_like(range_)
    o = np.ones_like(range_)

    # Allow for S and radius to be a single number
    if len(S) == 1:
        S = S * o
    if len(radius) == 1:
        radius = radius * o

    # Interpret some of the options
    nfov = len(rho_fov)
    is_hsrl = "-hsrl" in options
    is_adjoint = "-adjoint" in options
    is_jacobian = "-jacobian" in options

    # Open the input file and write the comments and the first line
    with open(in_temp, 'w') as fid:
        fid.write('# This file contains an input profile of atmospheric properties for\n')
        fid.write('# the multiscatter code. The comand-line should be of the form:\n')
        fid.write('#   ./multiscatter [options] [input_file] > [output_file]\n')
        fid.write('# or\n')
        fid.write('#   ./multiscatter [options] < [input_file] > [output_file]\n')
        if options:
            fid.write(f'# Suitable options for this input file are:\n#   {options}\n')
        else:
            fid.write('# No options are necessary for this particular file.\n')
        fid.write('# The file format consists of any number of comment lines starting\n')
        fid.write('# with "#", followed by a line of 5 inversion values:\n')
        fid.write('#   1. number of points in profile\n')
        fid.write('#   2. instrument wavelength (m)\n')
        fid.write('#   3. instrument altitude (m)\n')
        fid.write('#   4. transmitter 1/e half-width (radians)\n')
        fid.write('#  5+. receiver half-widths (radians)\n')
        fid.write('#      (1/e half-width if the "-gaussian-receiver" option\n')
        fid.write('#       is specified, top-hat half-width otherwise)\n')
        fid.write('# The subsequent lines contain 4 or more elements:\n')
        fid.write('#   1. range above ground, starting with nearest point to instrument (m)\n')
        fid.write('#   2. extinction coefficient of cloud/aerosol only (m-1)\n')
        fid.write('#   3. equivalent-area particle radius of cloud/aerosol (m)\n')
        fid.write('#   4. extinction-to-backscatter ratio of cloud/aerosol (sterad)\n')
        fid.write('#   5. extinction coefficient of air (m-1) (default 0)\n')
        fid.write('#   6. single scattering albedo of cloud/aerosol\n')
        fid.write('#   7. scattering asymmetry factor of cloud/aerosol\n')
        fid.write('#   8. single scattering albedo of air (isotropic scattering)\n')
        fid.write('#   9. fraction of cloud/aerosol backscatter due to droplets\n')
        fid.write('#  10. fraction of cloud/aerosol backscatter due to pristine ice\n')
        fid.write('# Note that elements 6-8 correspond to the wide-angle\n')
        fid.write('# multiple-scattering calculation and if this is omitted then only\n')
        fid.write('# the small-angle multiple-scattering calculation is performed.\n')
        fid.write('# For more help on how to run the code, type:\n')
        fid.write('#   ./multiscatter -help\n')
        fid.write(f'{len(range_)} {wavelength} {alt} {rho_div:g}')
        for rho in rho_fov:
            fid.write(f' {rho:g}')
        fid.write('\n')

        # Set variables that may be missing
        if nargin < 15:
            pristine_ice_fraction = z
            if nargin < 14:
                droplet_fraction = z
                if nargin < 13:
                    if wavelength > 1e-6:
                        ssa_air = z
                    else:
                        ssa_air = o

        if nargin >= 12:
            # Allow g, ssa, ssa_air and *_fraction to be single values
            if len(g) == 1:
                g = g * o
            if len(ssa) == 1:
                ssa = ssa * o
            if len(ssa_air) == 1:
                ssa_air = ssa_air * o
            if len(droplet_fraction) == 1:
                droplet_fraction = droplet_fraction * o
            if len(pristine_ice_fraction) == 1:
                pristine_ice_fraction = pristine_ice_fraction * o
            if nargin > 15:  # tem problema
                # Adjoint
                if not is_hsrl:
                    bscat_air_AD = np.zeros(bscat_AD.shape)
                for ii in range(len(range_)):
                    data = [range_[ii], ext[ii], radius[ii], S[ii], ext_air[ii], ssa[ii], g[ii],
                            ssa_air[ii], droplet_fraction[ii], pristine_ice_fraction[ii],
                            *bscat_AD[ii, :], *bscat_air_AD[ii, :]]
                    line = ' '.join([str(d) for d in data]) + '\n'
                    fid.write(line)
            else:
                # Wide-angle calculation
                for z in range(len(range_)):
                    line = "{:.0f} {:.8e} {:.6f} {:.6f} {:.8e} {:.6f} {:.1f} {:.8e} {:.0f} {:.0f}\n".format(
                        range_[z], ext[z], radius[z], S[z], ext_air[z], ssa[z], g[z], ssa_air[z], droplet_fraction[z],
                        pristine_ice_fraction[z]
                    )
                    fid.write(line)
        elif nargin >= 10:  # tem problema
            # Small-angle calculation only, with specified molecular extinction
            fid.write(' '.join(map(str, [range_, ext, radius, S, ext_air])))
            fid.write('\n')
        else:  # tem problema
            # Small-angle calculation only, with no molecular extinction
            fid.write(' '.join(map(str, [range_, ext, radius, S])))
            fid.write('\n')

    # Run executable and read output
    command = [
        'powershell.exe', '-Command',
        f"Get-Content {in_temp} | & 'C:\\Users\\luant\\PycharmProjects\\lidarpy\\lidarpy\\multiple_scattering"
        f"\\multiscatter.exe' {options} 2> {out_temp}"
    ]

    # executar o comando e capturar saída
    output, error = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()

    status = 0 if error.decode() == '' else 1

    if status != 0:
        print(error)
        with open(out_temp, 'r') as f:
            print(f.read())
        raise Exception('An error occurred in running the executable')

    try:
        output = output.decode()
    except Exception as e:
        print(error)
        with open(out_temp, 'r') as f:
            print(f.read())
        raise Exception('Problem interpreting output as a table of numbers')

    if not is_jacobian:
        # Create a dictionary containing the output variables
        data = {'range': [], 'ext': [], 'radius': [], 'bscat': []}
        for line in output.split('\n'):
            if line == "":
                continue
            items = line.split()
            data['range'].append(float(items[1]))
            data['ext'].append(float(items[2]))
            data['radius'].append(float(items[3]))
            data['bscat'].append(float(items[4]))

        data['range'] = np.array(data['range'])
        data['ext'] = np.array(data['ext'])
        data['radius'] = np.array(data['radius'])
        data['bscat'] = np.array(data['bscat'])

        index = 4 + nfov
        if is_hsrl:
            data['bscat_air'] = [[row[z] for row in output] for _ in range(index, index + nfov)]
            index += nfov
        if is_adjoint:
            data['ext_AD'] = [row[index + 1] for row in output]
            data['ssa_AD'] = [row[index + 2] for row in output]
            data['g_AD'] = [row[index + 3] for row in output]
            data['ext_bscat_ratio_AD'] = [row[index + 4] for row in output]
    else:
        n = len(range_)
        if is_hsrl:
            data = {'d_bscat_d_ext': [row[:len(row) // 2] for row in output[:n]],
                    'd_bscat_air_d_ext': [row[len(row) // 2:] for row in output[:n]]}
        else:
            data = {'d_bscat_d_ext': [row for row in output[:n]]}

        if len(output) > n:
            if is_hsrl:
                data['d_bscat_d_ssa'] = [row[:len(row) // 2] for row in output[n:n * 2]]
                data['d_bscat_air_d_ssa'] = [row[len(row) // 2:] for row in output[n:n * 2]]
                data['d_bscat_d_g'] = [row[:len(row) // 2] for row in output[2 * n:2 * n + n]]
                data['d_bscat_air_d_g'] = [row[len(row) // 2:] for row in output[2 * n:2 * n + n]]
                data['d_bscat_d_radius'] = [row[:len(row) // 2] for row in output[3 * n:3 * n + n]]
                data['d_bscat_air_d_radius'] = [row[len(row) // 2:] for row in output[3 * n:3 * n + n]]
                data['d_bscat_d_ext_bscat_ratio'] = output[4 * n + 1][:len(output[4 * n + 1]) // 2]
            else:
                data['d_bscat_d_ssa'] = output[n:n * 2, :]
                data['d_bscat_d_g'] = output[2 * n:2 * n + n, :]
                data['d_bscat_d_radius'] = output[3 * n:3 * n + n, :]
                data['d_bscat_d_ext_bscat_ratio'] = output[4 * n, :]

    # delete temporary files
    os.remove(in_temp)
    os.remove(out_temp)

    return data




import xarray as xr
import numpy as np
from lidarpy.utils.functions import molecular_model, z_finder
from lidarpy.inversion import Klett
from scipy.integrate import trapz
from scipy.interpolate import interp1d


class GetCod:
    """
    Implements a method for determining Cloud Optical Depth (COD) using lidar inversion.

    This class applies techniques described in molecular scattering and radiative transfer theory to calculate the cloud optical depth.
    The primary methods allow for both standard fitting and Monte Carlo techniques for uncertainty estimation.

    Attributes:
    - lidar_data (xr.Dataset): The dataset containing the lidar signals.
    - rangebin (array-like): Range or altitude bins of the lidar inversion.
    - wavelength (int): Emission wavelength of the lidar in nanometers.
    - molecular_data (xr.Dataset): The dataset containing the molecular or Rayleigh scattered signals.
    - pc (bool): Flag indicating if polarization correction should be applied.
    - co2ppmv (int): Concentration of CO2 in the atmosphere in ppmv.
    - fit_ref (list): Range indices for fitting the molecular profile.
    - transmittance_ref (int): Index for the reference altitude for determining transmittance.
    - mc_iter (int or None): Number of Monte Carlo iterations for uncertainty estimation.
    - indx (int): Index for a specific profile or inversion subset.
    - fit_ (bool): If True, it will fit the inversion.
    - delta_fit (int or None): Height interval for transmission fitting.
    - mc_niter (int): Number of monte carlo interations

    Methods:
    - fit: Calculates the cloud optical depth and the associated uncertainties using either the standard or Monte Carlo method.
    """

    _fit_delta_z: float = 3000
    _delta_z: float = 100
    _transmittance_bins: int = 150

    def __init__(self, rangebin: np.array, signal: np.array, molecular_data: xr.Dataset, cloud_lims: list,
                 fit_region: list):
        self.signal = signal
        self.rangebin = rangebin
        self.molecular_data = molecular_data
        self.cloud_lims = cloud_lims
        self.fit_region = fit_region

    def set_fit_delta_z(self, fit_delta_z):
        self._fit_delta_z: float = fit_delta_z
        return self

    def set_delta_z(self, delta_z):
        self._delta_z: float = delta_z
        return self

    def set_transmittance_bins(self, transmittance_bins):
        self._transmittance_bins: float = transmittance_bins
        return self

    def fit(self):
        molecular_signal = molecular_model(self.rangebin, self.signal, self.molecular_data, self.fit_region)

        molecular_rcs = molecular_signal * self.rangebin ** 2

        rcs = self.signal * self.rangebin ** 2

        transmittance_ref = z_finder(self.rangebin, self.cloud_lims[1] + self._delta_z)
        transmittance_ = (rcs[transmittance_ref: transmittance_ref + self._transmittance_bins]
                          / molecular_rcs[transmittance_ref: transmittance_ref + self._transmittance_bins])

        mean_ = transmittance_.mean()
        std_ = transmittance_.std(ddof=1)
        std_mean = std_ / np.sqrt(len(transmittance_))

        cod_mean = -0.5 * np.log(mean_)
        cod_std = 0.5 * std_mean / mean_

        return cod_mean, cod_std


class LidarRatioCalculator:
    def __init__(self, rangebin: np.array, signal: np.array, molecular_data: xr.Dataset, fit_region: list,
                 molecular_reference_region: list, cloud_lims: list, mc=False, mc_niter=100,
                 correct_noise: bool = True):
        self.signal = signal
        self.rangebin = rangebin
        self.cloud_lims = cloud_lims
        self.molecular_data = molecular_data
        self.molecular_reference_region = molecular_reference_region
        self.correct_noise = correct_noise
        self.mc = mc
        self.mc_niter = mc_niter
        self.fit_region = fit_region
        self.tau_mean, self.tau_std_mean = None, None
        self.tau_transmittance = None

    def _get_lr(self, tau_transmittance):
        lidar_ratios = np.arange(2, 50, 1)

        cloud_ind = z_finder(self.rangebin, self.cloud_lims)

        klett = Klett(rangebin=self.rangebin, signal=self.signal, molecular_data=self.molecular_data, lidar_ratio=1,
                      molecular_reference_region=self.molecular_reference_region, correct_noise=self.correct_noise)

        taus = []
        for lidar_ratio in lidar_ratios:
            klett.set_lidar_ratio(lidar_ratio)
            alpha, *_ = klett.fit()
            taus.append(trapz(y=alpha[cloud_ind[0]:cloud_ind[1] + 1],
                              dx=self.rangebin[1] - self.rangebin[0]))

        difference = (np.array(taus) - tau_transmittance) ** 2

        f_diff = interp1d(lidar_ratios, difference, kind="cubic", fill_value="extrapolate")

        new_lr = np.linspace(8, 50, 100)

        new_diff = f_diff(new_lr)

        return new_lr[new_diff.argmin()]

    def _compute_tau(self):
        self.tau_mean, self.tau_std_mean = GetCod(rangebin=self.rangebin, signal=self.signal,
                                                  molecular_data=self.molecular_data, cloud_lims=self.cloud_lims,
                                                  fit_region=self.fit_region).fit()

        self.tau_transmittance = self.tau_mean + self.tau_std_mean * np.random.randn(self.mc_niter)

    def _has_valid_transmittance(self):
        valid_ratio = len(self.tau_transmittance[self.tau_transmittance >= 0]) / len(self.tau_transmittance)
        return valid_ratio > 0.5

    def _compute_valid_lrs(self):
        return [self._get_lr(tau_transmittance=tau) for tau in self.tau_transmittance if tau >= 0]

    @staticmethod
    def _get_lidar_ratio_values(lrs):
        valid_lrs = [lr for lr in lrs if 8.38 <= lr < 48]
        if len(valid_lrs) / len(lrs) >= 0.6:
            return np.mean(valid_lrs), np.std(valid_lrs, ddof=1)
        return 25, 0

    def fit(self):
        if self.tau_transmittance is None:
            self._compute_tau()

        if not self._has_valid_transmittance():
            return 25, 0, self.tau_mean, self.tau_std_mean

        if self.mc:
            lrs = self._compute_valid_lrs()
            lr, lr_error = self._get_lidar_ratio_values(lrs)
        else:
            lr = self._get_lr(tau_transmittance=self.tau_mean)
            lr_error = 0

        return lr, lr_error, self.tau_mean, self.tau_std_mean



from setuptools import setup, find_packages

VERSION = '0.0.9'
DESCRIPTION = 'LIDARpy is a comprehensive Python library tailored for the analysis, manipulation, and interpretation of LIDAR data.'
LONG_DESCRIPTION = 'LIDARpy is a comprehensive Python library tailored for the analysis, manipulation, and interpretation of LIDAR data. This library provides a set of tools for background noise removal, data grouping, bin adjustments, uncertainty computations, and advanced data inversion using both the Klett and Raman methods.'

# Setting up
setup(
    name="lidarpy",
    version=VERSION,
    author="Luan Cordeiro",
    author_email="<luancordeiro@usp.br>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['numpy', 'xarray', 'scipy', 'datetime'],
    keywords=['python', 'lidar', 'clouds', 'data'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



import numpy as np
import xarray as xr
from lidarpy.utils.constants import Constants


class AlphaBetaMolecular:
    """
    Computes molecular optical properties based on atmospheric conditions.

    Given an atmospheric pressure profile, temperature profile, wavelength,
    and CO2 concentration, this class provides methods to calculate the molecular
    extinction coefficient, molecular backscattering coefficient, and molecular lidar ratio.

    Parameters:
    -----------
    p_air : np.ndarray
        Atmospheric pressure profile [P].
    t_air : np.ndarray
        Atmospheric temperature profile [K].
    wavelength : float
        Wavelength in nanometers.
    co2ppmv : float
        CO2 concentration in parts per million by volume.

    Methods:
    --------
    set_wavelength(wavelength: float) -> AlphaBetaMolecular:
        Set a new wavelength for calculations.

    get_params() -> Tuple[np.ndarray, np.ndarray, float]:
        Compute and return the molecular extinction coefficient, molecular backscattering coefficient,
        and molecular lidar ratio.

    Attributes:
    -----------
    alpha_mol : np.ndarray
        Molecular extinction coefficient [m^-1].
    beta_mol : np.ndarray
        Molecular backscattering coefficient [m^-1 sr^-1].
    lr_mol : float
        Molecular lidar ratio [sr].

    Authors:
    --------
    H. M. J. Barbosa (hbarbosa@if.usp.br), IF, USP, Brazil
    B. Barja (bbarja@gmail.com), GOAC, CMC, Cuba
    R. Costa (re.dacosta@gmail.com), IPEN, Brazil
    """

    def __init__(self, rangebin: np.ndarray, p_air: np.ndarray, t_air: np.ndarray, wavelength: float,
                 co2ppmv: float = 372):
        self.rangebin = rangebin
        self.p_air = p_air
        self.t_air = t_air
        self.wavelength = wavelength * 1e-9
        self.const = Constants(co2ppmv)

    def set_wavelength(self, wavelength):
        self.wavelength = wavelength * 1e-9
        return self

    def _refractive_index(self):
        if self.wavelength * 1e6 > 0.23:
            dn300 = (5791817 / (238.0185 - 1 / (self.wavelength * 1e6) ** 2) +
                     167909 / (57.362 - 1 / (self.wavelength * 1e6) ** 2)) * 1e-8
        else:
            dn300 = (8060.51 + 2480990 / (132.274 - 1 / (self.wavelength * 1e6) ** 2) +
                     14455.7 / (39.32957 - 1. / (self.wavelength * 1e6) ** 2)) * 1e-8

        dn_air = dn300 * (1 + (0.54 * (self.const.co2ppmv * 1e-6 - 0.0003)))
        n_air = dn_air + 1
        return n_air

    def _king_factor(self):
        f_n2 = 1.034 + (3.17e-4 / ((self.wavelength * 1e6) ** 2))
        f_o2 = 1.096 + (1.385e-3 / ((self.wavelength * 1e6) ** 2)) + (1.448e-4 / ((self.wavelength * 1e6) ** 4))
        f_ar = 1
        f_co2 = 1.15
        f_air = (self.const.N2ppv * f_n2 + self.const.O2ppv * f_o2 + self.const.Arppv * f_ar +
                 self.const.co2ppmv * 1e-6 * f_co2) / (self.const.N2ppv + self.const.O2ppv + self.const.Arppv +
                                                       self.const.co2ppmv * 1e-6)
        return f_air

    def _depolarization_ratio(self):
        f_air = self._king_factor()
        rho_air = (6 * f_air - 6) / (3 + 7 * f_air)
        return rho_air

    def _phase_function(self):
        rho_air = self._depolarization_ratio()
        gamma_air = rho_air / (2 - rho_air)
        pf_mol = 0.75 * ((1 + 3 * gamma_air) + (1 - gamma_air) * (np.cos(np.pi) ** 2)) / (1 + 2 * gamma_air)
        return pf_mol

    def _cross_section(self):
        n_air = self._refractive_index()
        f_air = self._king_factor()
        sigma_std = 24 * (np.pi ** 3) * ((n_air ** 2 - 1) ** 2) * f_air / ((self.wavelength ** 4) *
                                                                           (self.const.Nstd ** 2) *
                                                                           (((n_air ** 2) + 2) ** 2))
        return sigma_std

    def _vol_scattering_coeff(self):
        sigma_std = self._cross_section()
        alpha_std = self.const.Nstd * sigma_std
        alpha_mol = self.p_air / self.t_air * self.const.Tstd / self.const.Pstd * alpha_std
        return alpha_mol

    def _ang_vol_scattering_coeff(self, alpha_mol):
        pf_mol = self._phase_function()
        lr_mol = 4 * np.pi / pf_mol
        beta_mol = alpha_mol / lr_mol
        return beta_mol, lr_mol

    def get_params(self):
        alpha_mol = self._vol_scattering_coeff()
        beta_mol, lr_mol = self._ang_vol_scattering_coeff(alpha_mol)

        alpha_mol = xr.DataArray(alpha_mol, coords=[('rangebin', self.rangebin)], name='alpha')
        beta_mol = xr.DataArray(beta_mol, coords=[('rangebin', self.rangebin)], name='beta')
        lr_mol = xr.DataArray(lr_mol, coords=[('rangebin', self.rangebin)], name='lidar_ratio')

        ds = xr.merge([alpha_mol, beta_mol, lr_mol])

        return ds
